from asr.core import command, option

test1 = {'description': 'Test ground state of Si.',
         'cli': ['asr run "setup.materials -s Si2"',
                 'ase convert materials.json structure.json',
                 'asr run "setup.params *:calculator '
                 "{'name':'gpaw','mode':'lcao','kpts':(4,4,4),...}" '"',
                 'asr run gs@calculate',
                 'asr run database.fromtree',
                 'asr run "database.browser --only-figures"']}


@command(module='asr.gs',
         creates=['gs.gpw'],
         tests=[test1],
         requires=['structure.json'],
         resources='8:10h')
@option('-c', '--calculator', help='Calculator params.')
def calculate(calculator={'name': 'gpaw',
                          'mode': {'name': 'pw', 'ecut': 800},
                          'xc': 'PBE',
                          'basis': 'dzp',
                          'kpts': {'density': 12.0, 'gamma': True},
                          'occupations': {'name': 'fermi-dirac',
                                          'width': 0.05},
                          'convergence': {'bands': 'CBM+3.0'},
                          'nbands': '200%',
                          'txt': 'gs.txt',
                          'charge': 0}):
    """Calculate ground state file.

    This recipe saves the ground state to a file gs.gpw based on the structure
    in 'structure.json'. This can then be processed by asr.gs@postprocessing
    for storing any derived quantities. See asr.gs@postprocessing for more
    information.
    """
    import numpy as np
    from ase.io import read
    from ase.calculators.calculator import PropertyNotImplementedError
    atoms = read('structure.json')

    nd = np.sum(atoms.pbc)
    if nd == 2:
        assert not atoms.pbc[2], \
            'The third unit cell axis should be aperiodic for a 2D material!'
        calculator['poissonsolver'] = {'dipolelayer': 'xy'}

    from ase.calculators.calculator import get_calculator_class
    name = calculator.pop('name')
    calc = get_calculator_class(name)(**calculator)

    atoms.set_calculator(calc)
    atoms.get_forces()
    try:
        atoms.get_stress()
    except PropertyNotImplementedError:
        pass
    atoms.get_potential_energy()
    atoms.calc.write('gs.gpw')


tests = [{'description': 'Test ground state of Si.',
          'tags': ['gitlab-ci'],
          'cli': ['asr run "setup.materials -s Si2"',
                  'asr run "setup.params *:calculator '
                  '''{'name':'gpaw','mode':'lcao','kpts':(4,4,4)}"''',
                  'ase convert materials.json structure.json',
                  'asr run gs',
                  'asr run database.fromtree',
                  'asr run "database.browser --only-figures"']}]


def webpanel(row, key_descriptions):
    from asr.database.browser import table, fig

    t = table(row, 'Property',
              ['gap', 'gap_dir',
               'dipz', 'evacdiff', 'workfunction', 'dos_at_ef_soc'],
              key_descriptions)

    gap = row.get('gap')

    if gap > 0:
        if row.get('evac'):
            t['rows'].extend(
                [['Valence band maximum wrt. vacuum level',
                  f'{row.vbm - row.evac:.2f} eV'],
                 ['Conduction band minimum wrt. vacuum level',
                  f'{row.cbm - row.evac:.2f} eV']])
        else:
            t['rows'].extend(
                [['Valence band maximum wrt. Fermi level',
                  f'{row.vbm - row.efermi:.2f} eV'],
                 ['Conduction band minimum wrt. Fermi level',
                  f'{row.cbm - row.efermi:.2f} eV']])
    panel = {'title': 'Basic electronic properties (PBE)',
             'columns': [[t], [fig('bz-with-gaps.png')]],
             'sort': 10}

    row = ['Band gap (PBE)', f'{row.gap:0.2f} eV']
    summary = {'title': 'Summary',
               'columns': [[{'type': 'table',
                             'header': ['Electronic properties', ''],
                             'rows': [row]}]],
               'plot_descriptions': [{'function': bz_soc,
                                      'filenames': ['bz-with-gaps.png']}],
               'sort': 10}

    return [panel, summary]


def bz_soc(row, fname):
    from ase.geometry.cell import Cell
    from matplotlib import pyplot as plt
    import numpy as np
    cell = Cell(row.cell)
    lat = cell.get_bravais_lattice(pbc=row.pbc)
    plt.figure(figsize=(4, 4))
    lat.plot_bz(vectors=False, pointstyle={'c': 'k', 'marker': '.'})
    gsresults = row.data.get('results-asr.gs.json')
    cbm_c = gsresults['k_cbm_c']
    vbm_c = gsresults['k_vbm_c']

    if cbm_c is not None:
        ax = plt.gca()
        icell = np.linalg.inv(row.cell).T
        cbm_v = np.dot(cbm_c, icell)
        vbm_v = np.dot(vbm_c, icell)

        vbm_style = {'marker': 'o', 'facecolor': 'w',
                     'edgecolors': 'C0', 's': 100, 'lw': 2,
                     'zorder': 4}
        cbm_style = {'c': 'C1', 'marker': 'o', 's': 40, 'zorder': 5}
        ax.scatter([vbm_v[0]], [vbm_v[1]], **vbm_style, label='VBM')
        ax.scatter([cbm_v[0]], [cbm_v[1]], **cbm_style, label='CBM')
        xlim = np.array(ax.get_xlim()) * 1.4
        ylim = np.array(ax.get_ylim()) * 1.4
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        plt.legend(loc='upper center', ncol=3)

    plt.tight_layout()
    plt.savefig(fname)


@command(module='asr.gs',
         requires=['gs.gpw', 'structure.json',
                   'results-asr.magnetic_anisotropy.json'],
         tests=tests,
         dependencies=['asr.gs@calculate', 'asr.magnetic_anisotropy'],
         webpanel=webpanel)
def main():
    """Extract derived quantities from groundstate in gs.gpw."""
    import numpy as np
    from ase.io import read
    from asr.calculators import get_calculator
    from gpaw.mpi import serial_comm

    # Just some quality control before we start
    atoms = read('structure.json')
    calc = get_calculator()('gs.gpw', txt=None,
                            communicator=serial_comm)
    pbc = atoms.pbc
    ndim = np.sum(pbc)

    if ndim == 2:
        assert not pbc[2], \
            'The third unit cell axis should be aperiodic for a 2D material!'
        # For 2D materials we check that the calculater used a dipole
        # correction if the material has an out-of-plane dipole

        # Small hack
        atoms = calc.atoms
        atoms.calc = calc
        evacdiffmin = 10e-3
        if evacdiff(calc.atoms) > evacdiffmin:
            assert calc.todict().get('poissonsolver', {}) == \
                {'dipolelayer': 'xy'}, \
                ('The ground state has a finite dipole moment along aperiodic '
                 'axis but calculation was without dipole correction.')

    # Now that some checks are done, we can extract information
    forces = calc.get_property('forces', allow_calculation=False)
    stresses = calc.get_property('stress', allow_calculation=False)
    etot = calc.get_potential_energy()

    results = {'forces': forces,
               'stresses': stresses,
               'etot': etot}

    results['gaps_nosoc'] = gaps(calc, soc=False)
    results['gap_dir_nosoc'] = results['gaps_nosoc']['gap_dir']
    results.update(gaps(calc, soc=True))
    # Vacuum level is calculated for c2db backwards compability
    if int(np.sum(atoms.get_pbc())) == 2:
        vac = vacuumlevels(atoms, calc)
        results['vacuumlevels'] = vac
        results['dipz'] = vac['dipz']
        results['evac'] = vac['evacmean']
        results['evacdiff'] = vac['evacdiff']
        results['workfunction'] = results['evac'] - results['efermi']

    fingerprint = {}
    for setup in calc.setups:
        fingerprint[setup.symbol] = setup.fingerprint
    results['__setup_fingerprints__'] = fingerprint
    results['__key_descriptions__'] = {
        # Saved arrays
        'forces': 'Forces on atoms [eV/Angstrom]',
        'stresses': 'Stress on unit cell [eV/Angstrom^dim]',
        # Key value pairs
        'etot': 'KVP: Total energy (Tot. En.) [eV]',
        'evac': 'KVP: Vacuum level (Vacuum level) [eV]',
        'evacdiff': 'KVP: Vacuum level shift (Vacuum level shift) [eV]',
        'dipz': 'KVP: Out-of-plane dipole [e * Ang]',
        'efermi': 'KVP: Fermi level (Fermi level) [eV]',
        'gap': 'KVP: Band gap (Band gap) [eV]',
        'vbm': 'KVP: Valence band maximum (Val. band max.) [eV]',
        'cbm': 'KVP: Conduction band minimum (Cond. band max.) [eV]',
        'gap_dir': 'KVP: Direct band gap (Dir. band gap) [eV]',
        'vbm_dir': ('KVP: Direct valence band maximum '
                    '(Dir. val. band max.) [eV]'),
        'cbm_dir': ('KVP: Direct conduction band minimum '
                    '(Dir. cond. band max.) [eV]'),
        'gap_dir_nosoc': ('KVP: Direct gap without SOC '
                          '(Dir. gap wo. soc.) [eV]')}

    return results


def gaps(calc, soc=True):
    # ##TODO min kpt dens? XXX
    # inputs: gpw groundstate file, soc?, direct gap? XXX
    from functools import partial
    from asr.utils.gpw2eigs import calc2eigs
    from asr.magnetic_anisotropy import get_spin_axis

    ibzkpts = calc.get_ibz_k_points()

    (evbm_ecbm_gap,
     skn_vbm, skn_cbm) = get_gap_info(soc=soc, direct=False,
                                      calc=calc)
    (evbm_ecbm_direct_gap,
     direct_skn_vbm, direct_skn_cbm) = get_gap_info(soc=soc, direct=True,
                                                    calc=calc)

    k_vbm, k_cbm = skn_vbm[1], skn_cbm[1]
    direct_k_vbm, direct_k_cbm = direct_skn_vbm[1], direct_skn_cbm[1]

    get_kc = partial(get_1bz_k, ibzkpts, calc)

    k_vbm_c = get_kc(k_vbm)
    k_cbm_c = get_kc(k_cbm)
    direct_k_vbm_c = get_kc(direct_k_vbm)
    direct_k_cbm_c = get_kc(direct_k_cbm)

    if soc:
        theta, phi = get_spin_axis()
        _, efermi = calc2eigs(calc, ranks=[0], soc=True,
                              theta=theta, phi=phi)
    else:
        efermi = calc.get_fermi_level()

    subresults = {'gap': evbm_ecbm_gap[2],
                  'vbm': evbm_ecbm_gap[0],
                  'cbm': evbm_ecbm_gap[1],
                  'gap_dir': evbm_ecbm_direct_gap[2],
                  'vbm_dir': evbm_ecbm_direct_gap[0],
                  'cbm_dir': evbm_ecbm_direct_gap[1],
                  'k_vbm_c': k_vbm_c,
                  'k_cbm_c': k_cbm_c,
                  'k_vbm_dir_c': direct_k_vbm_c,
                  'k_cbm_dir_c': direct_k_cbm_c,
                  'skn1': skn_vbm,
                  'skn2': skn_cbm,
                  'skn1_dir': direct_skn_vbm,
                  'skn2_dir': direct_skn_cbm,
                  'efermi': efermi}

    return subresults


def get_1bz_k(ibzkpts, calc, k_index):
    from gpaw.kpt_descriptor import to1bz
    k_c = ibzkpts[k_index] if k_index is not None else None
    if k_c is not None:
        k_c = to1bz(k_c[None], calc.wfs.gd.cell_cv)[0]
    return k_c


def get_gap_info(soc, direct, calc):
    from ase.dft.bandgap import bandgap
    from asr.utils.gpw2eigs import calc2eigs
    from asr.magnetic_anisotropy import get_spin_axis
    # e1 is VBM, e2 is CBM
    if soc:
        theta, phi = get_spin_axis()
        e_km, efermi = calc2eigs(calc, ranks=[0],
                                 soc=True, theta=theta, phi=phi)
        # km1 is VBM index tuple: (s, k, n), km2 is CBM index tuple: (s, k, n)
        gap, km1, km2 = bandgap(eigenvalues=e_km, efermi=efermi, direct=direct,
                                kpts=calc.get_ibz_k_points(), output=None)
        if km1[0] is not None:
            e1 = e_km[km1]
            e2 = e_km[km2]
        else:
            e1, e2 = None, None
        x = (e1, e2, gap), (0,) + tuple(km1), (0,) + tuple(km2)
    else:
        g, skn1, skn2 = bandgap(calc, direct=direct, output=None)
        if skn1[1] is not None:
            e1 = calc.get_eigenvalues(spin=skn1[0], kpt=skn1[1])[skn1[2]]
            e2 = calc.get_eigenvalues(spin=skn2[0], kpt=skn2[1])[skn2[2]]
        else:
            e1, e2 = None, None
        x = (e1, e2, g), skn1, skn2
    return x


def vacuumlevels(atoms, calc, n=8):
    """Get the vacuumlevels on both sides of a 2D material.

    Get the vacuumlevels on both sides of a 2D material. Will
    do a dipole corrected dft calculation, if needed (Janus structures).
    Assumes the 2D material periodic directions are x and y.
    Assumes that the 2D material is centered in the z-direction of
    the unit cell.

    Dipole corrected dft calculation -> dipcorrgs.gpw

    Parameters
    ----------
    gpw: str
       name of gpw file to base the dipole corrected calc on
    evacdiffmin: float
        thresshold in eV for doing a dipole moment corrected
        dft calculations if the predicted evac difference is less
        than this value don't do it
    n: int
        number of gridpoints away from the edge to evaluate the vac levels
    """
    import numpy as np

    # Record electrostatic potential as a function of z
    v_z = calc.get_electrostatic_potential().mean(0).mean(0)
    z_z = np.linspace(0, atoms.cell[2, 2], len(v_z), endpoint=False)

    # Store data
    subresults = {'z_z': z_z, 'v_z': v_z,
                  'evacdiff': evacdiff(atoms),
                  'dipz': atoms.get_dipole_moment()[2],
                  # Extract vaccuum energy on both sides of the slab
                  'evac1': v_z[n],
                  'evac2': v_z[-n],
                  'evacmean': (v_z[n] + v_z[-n]) / 2,
                  # Ef might have changed in dipole corrected gs
                  'efermi_nosoc': calc.get_fermi_level()}

    return subresults


def evacdiff(atoms):
    """Derive vacuum energy level difference from the dipole moment.

    Calculate vacuum energy level difference from the dipole moment of
    a slab assumed to be in the xy plane

    Returns
    -------
    out: float
        vacuum level difference in eV
    """
    import numpy as np
    from ase.units import Bohr, Hartree

    A = np.linalg.det(atoms.cell[:2, :2] / Bohr)
    dipz = atoms.get_dipole_moment()[2] / Bohr
    evacsplit = 4 * np.pi * dipz / A * Hartree

    return evacsplit


if __name__ == '__main__':
    main.cli()



import pytest
from ase.build import bulk

Si = bulk('Si')


@pytest.mark.ci
def test_setup_decorate_si(separate_folder, mockgpaw):
    from asr.setup.decorate import main
    from ase.io import write
    from ase.db import connect
    from pathlib import Path
    write('structure.json', Si)
    main(atoms='structure.json')
    assert Path('decorated.db').is_file()
    db = connect('decorated.db')
    db.get('Ge')



from asr.core import command, option

from collections import defaultdict

import numpy as np

from ase import Atoms
from ase.units import Ha
from ase.dft.kpoints import get_monkhorst_pack_size_and_offset as k2so
from ase.dft.dos import DOS
from ase.dft.dos import linear_tetrahedron_integration as lti

from ase.units import Hartree

from asr.core import magnetic_atoms, read_json


# ---------- GPAW hacks ---------- #


# Hack the density of states
class SOCDOS(DOS):
    def __init__(self, gpw, npts=401, **kwargs):
        """Hack to make DOS class work with spin orbit coupling.

        Parameters
        ----------
        gpw : str
            The SOCDOS takes a filename of the GPAW calculator object and loads
            it, instead of the normal ASE compliant calculator object.
        npts : int
            see ase.dft.dos.DOS

        """
        # Initiate DOS with serial communicator instead
        from gpaw import GPAW
        import gpaw.mpi as mpi
        from asr.utils.gpw2eigs import calc2eigs
        from asr.magnetic_anisotropy import get_spin_axis

        # Initiate calculator object and get the spin-orbit eigenvalues
        calc = GPAW(gpw, communicator=mpi.serial_comm, txt=None)
        theta, phi = get_spin_axis()
        e_skm, ef = calc2eigs(calc, theta=theta, phi=phi, ranks=[0])

        # Only the rank=0 should have an actual DOS object.
        # The others receive the output as a broadcast.
        self.world = mpi.world
        if mpi.world.rank == 0:
            DOS.__init__(self, calc, npts=npts, **kwargs)

            # Hack the number of spins
            self.nspins = 1

            # Hack the eigenvalues
            if e_skm.ndim == 2:
                e_skm = e_skm[np.newaxis]
            e_skn = e_skm - ef
            bzkpts = calc.get_bz_k_points()
            size, offset = k2so(bzkpts)
            bz2ibz = calc.get_bz_to_ibz_map()
            shape = (self.nspins, ) + tuple(size) + (-1, )
            self.e_skn = e_skn[:, bz2ibz].reshape(shape)
        else:
            self.npts = npts

    def get_dos(self):
        """Interface to DOS.get_dos."""
        # Rank=0 calculates the dos
        if self.world.rank == 0:
            dos = np.ascontiguousarray(DOS.get_dos(self, spin=0))
        else:
            dos = np.empty(self.npts)

        # Broadcast result
        self.world.broadcast(dos, 0)

        return dos


# Hack the local density of states to keep spin-orbit results and not
# compute them repeatedly
class SOCDescriptor:
    """Descriptor for spin-orbit corrections.

    [Developed and tested for raw_spinorbit_orbital_LDOS only]
    """

    def __init__(self, paw):
        self.paw = paw

        # Log eigenvalues and wavefunctions for multiple spin directions
        self.theta_d = []
        self.phi_d = []
        self.eps_dmk = []
        self.v_dknm = []

    def calculate_soc_eig(self, theta, phi):
        from gpaw.spinorbit import get_spinorbit_eigenvalues
        eps_mk, v_knm = get_spinorbit_eigenvalues(self.paw, return_wfs=True,
                                                  theta=theta, phi=phi)
        self.theta_d.append(theta)
        self.phi_d.append(phi)
        self.eps_dmk.append(eps_mk)
        self.v_dknm.append(v_knm)

    def get_soc_eig(self, theta=0, phi=0):
        # Check if eigenvalues have been computed already
        for d, (t, p) in enumerate(zip(self.theta_d, self.phi_d)):
            if abs(t - theta) < np.pi * 1.e-4 and abs(p - phi) < np.pi * 1.e-4:
                return self.eps_dmk[d], self.v_dknm[d]
        # Calculate if not
        self.calculate_soc_eig(theta, phi)
        return self.eps_dmk[-1], self.v_dknm[-1]


def raw_spinorbit_orbital_LDOS_hack(paw, a, spin, angular='spdf',
                                    theta=0, phi=0):
    """Hack raw_spinorbit_orbital_LDOS."""
    from gpaw.utilities.dos import get_angular_projectors
    from gpaw.spinorbit import get_spinorbit_projections

    # Attach SOCDescriptor to the calculator object
    if not hasattr(paw, 'socd'):
        paw.socd = SOCDescriptor(paw)

    # Get eigenvalues and wavefunctions from SOCDescriptor
    eps_mk, v_knm = paw.socd.get_soc_eig(theta, phi)
    e_mk = eps_mk / Hartree

    # Do the rest as usual:
    ns = paw.wfs.nspins
    w_k = paw.wfs.kd.weight_k
    nk = len(w_k)
    nb = len(e_mk)

    if a < 0:
        # Allow list-style negative indices; we'll need the positive a for the
        # dictionary lookup later
        a = len(paw.wfs.setups) + a

    setup = paw.wfs.setups[a]
    energies = np.empty(nb * nk)
    weights_xi = np.empty((nb * nk, setup.ni))
    x = 0
    for k, w in enumerate(w_k):
        energies[x:x + nb] = e_mk[:, k]
        P_ami = get_spinorbit_projections(paw, k, v_knm[k])
        if ns == 2:
            weights_xi[x:x + nb, :] = w * np.absolute(P_ami[a][:, spin::2])**2
        else:
            weights_xi[x:x + nb, :] = w * np.absolute(P_ami[a][:, 0::2])**2 / 2
            weights_xi[x:x + nb, :] += w * np.absolute(P_ami[a][:,
                                                                1::2])**2 / 2
        x += nb

    if angular is None:
        return energies, weights_xi
    elif isinstance(angular, int):
        return energies, weights_xi[:, angular]
    else:
        projectors = get_angular_projectors(setup, angular, type='bound')
        weights = np.sum(np.take(weights_xi,
                                 indices=projectors, axis=1), axis=1)
        return energies, weights


# ---------- Recipe tests ---------- #

params = "{'mode':{'ecut':200,...},'kpts':{'density':2.0},...}"
ctests = []
ctests.append({'description': 'Test the refined ground state of Si',
               'name': 'test_asr.pdos_Si_gpw',
               'tags': ['gitlab-ci'],
               'cli': ['asr run "setup.materials -s Si2"',
                       'ase convert materials.json structure.json',
                       'asr run "setup.params '
                       f'asr.gs@calculate:calculator {params} '
                       'asr.pdos@calculate:kptdensity 3.0 '
                       'asr.pdos@calculate:emptybands 5"',
                       'asr run gs',
                       'asr run pdos@calculate',
                       'asr run database.fromtree',
                       'asr run "database.browser --only-figures"']})

tests = []
tests.append({'description': 'Test the pdos of Si (cores=1)',
              'name': 'test_asr.pdos_Si_serial',
              'cli': ['asr run "setup.materials -s Si2"',
                      'ase convert materials.json structure.json',
                      'asr run "setup.params '
                      f'asr.gs@calculate:calculator {params} '
                      'asr.pdos@calculate:kptdensity 3.0 '
                      'asr.pdos@calculate:emptybands 5"',
                      'asr run gs',
                      'asr run pdos',
                      'asr run database.fromtree',
                      'asr run "database.browser --only-figures"']})
tests.append({'description': 'Test the pdos of Si (cores=2)',
              'name': 'test_asr.pdos_Si_parallel',
              'cli': ['asr run "setup.materials -s Si2"',
                      'ase convert materials.json structure.json',
                      'asr run "setup.params '
                      f'asr.gs@calculate:calculator {params} '
                      'asr.pdos@calculate:kptdensity 3.0 '
                      'asr.pdos@calculate:emptybands 5"',
                      'asr run gs',
                      'asr run -p 2 pdos',
                      'asr run database.fromtree',
                      'asr run "database.browser --only-figures"']})


# ---------- Webpanel ---------- #


def webpanel(row, key_descriptions):
    from asr.database.browser import fig, table
    # PDOS without spin-orbit coupling
    panel = {'title': 'Electronic band structure and projected DOS (PBE)',
             'columns': [[],
                         [fig('pbe-pdos_nosoc.png', link='empty')]],
             'plot_descriptions': [{'function': plot_pdos_nosoc,
                                    'filenames': ['pbe-pdos_nosoc.png']}],
             'sort': 14}

    # Another panel to make sure sorting is correct
    panel2 = {'title': 'Electronic band structure and projected DOS (PBE)',
              'columns': [[],
                          [table(row, 'Property', ['dos_at_ef_nosoc'],
                                 kd=key_descriptions)]]}

    return [panel, panel2]


# ---------- Main functionality ---------- #


# ----- Slow steps ----- #


@command(module='asr.pdos',
         creates=['pdos.gpw'],
         tests=ctests,
         requires=['gs.gpw'],
         dependencies=['asr.gs'])
@option('-k', '--kptdensity', type=float, help='K-point density')
@option('--emptybands', type=int, help='number of empty bands to include')
def calculate(kptdensity=20.0, emptybands=20):
    from asr.utils.refinegs import refinegs
    refinegs(selfc=False,
             kptdensity=kptdensity, emptybands=emptybands,
             gpw='pdos.gpw', txt='pdos.txt')


# ----- Fast steps ----- #


@command(module='asr.pdos',
         requires=['results-asr.gs.json', 'pdos.gpw'],
         tests=tests,
         dependencies=['asr.gs', 'asr.pdos@calculate'],
         webpanel=webpanel)
def main():
    from gpaw import GPAW
    from asr.core import singleprec_dict

    # Get refined ground state with more k-points
    calc = GPAW('pdos.gpw', txt=None)

    results = {}

    # Calculate the dos at the Fermi energy
    results['dos_at_ef_nosoc'] = dos_at_ef(calc, 'pdos.gpw', soc=False)
    results['dos_at_ef_soc'] = dos_at_ef(calc, 'pdos.gpw', soc=True)

    # Calculate pdos
    results['pdos_nosoc'] = singleprec_dict(pdos(calc, 'pdos.gpw', soc=False))
    results['pdos_soc'] = singleprec_dict(pdos(calc, 'pdos.gpw', soc=True))

    # Log key descriptions
    kd = {}
    kd['pdos_nosoc'] = ('Projected density of states '
                        'without spin-orbit coupling '
                        '(PDOS no soc)')
    kd['pdos_soc'] = ('Projected density of states '
                      'with spin-orbit coupling '
                      '(PDOS w. soc)')
    kd['dos_at_ef_nosoc'] = ('KVP: Density of states at the Fermi energy '
                             'without spin-orbit coupling '
                             '(DOS at ef no soc) [states/eV]')
    kd['dos_at_ef_soc'] = ('KVP: Density of states at the Fermi energy '
                           'with spin-orbit coupling '
                           '(DOS at ef w. soc) [states/eV]')
    results.update({'__key_descriptions__': kd})

    return results


# ---------- Recipe methodology ---------- #


# ----- PDOS ----- #


def pdos(calc, gpw, soc=True):
    """Do a single pdos calculation.

    Main functionality to do a single pdos calculation.
    """
    # Do calculation
    e_e, pdos_syl, symbols, ef = calculate_pdos(calc, gpw, soc=soc)

    return {'pdos_syl': pdos_syl, 'symbols': symbols,
            'energies': e_e, 'efermi': ef}


def calculate_pdos(calc, gpw, soc=True):
    """Calculate the projected density of states.

    Returns
    -------
    energies : nd.array
        energies 10 eV under and above Fermi energy
    pdos_syl : defaultdict
        pdos for spin, symbol and orbital angular momentum
    symbols : list
        chemical symbols in Atoms object
    efermi : float
        Fermi energy

    """
    from gpaw import GPAW
    import gpaw.mpi as mpi
    from gpaw.utilities.dos import raw_orbital_LDOS
    from gpaw.utilities.progressbar import ProgressBar
    from ase.utils import DevNull
    from ase.parallel import parprint
    from asr.magnetic_anisotropy import get_spin_axis
    world = mpi.world

    if soc and world.rank == 0:
        calc0 = GPAW(gpw, communicator=mpi.serial_comm, txt=None)

    zs = calc.atoms.get_atomic_numbers()
    chem_symbols = calc.atoms.get_chemical_symbols()
    efermi = calc.get_fermi_level()
    l_a = get_l_a(zs)
    kd = calc.wfs.kd

    if soc:
        ldos = raw_spinorbit_orbital_LDOS_hack
    else:
        ldos = raw_orbital_LDOS

    ns = calc.get_number_of_spins()
    theta, phi = get_spin_axis()
    gaps = read_json('results-asr.gs.json').get('gaps_nosoc')
    e1 = gaps.get('vbm') or gaps.get('efermi')
    e2 = gaps.get('cbm') or gaps.get('efermi')
    e_e = np.linspace(e1 - 3, e2 + 3, 500)

    # We distinguish in (spin(s), chemical symbol(y), angular momentum (l)),
    # that is if there are multiple atoms in the unit cell of the same chemical
    # species, their pdos are added together.
    pdos_syl = defaultdict(float)
    # Set up progressbar
    s_i = [s for s in range(ns) for a in l_a for l in l_a[a]]
    a_i = [a for s in range(ns) for a in l_a for l in l_a[a]]
    l_i = [l for s in range(ns) for a in l_a for l in l_a[a]]
    sal_i = [(s, a, l) for (s, a, l) in zip(s_i, a_i, l_i)]
    parprint('\nComputing pdos %s' % ('with spin-orbit coupling' * soc))
    if mpi.world.rank == 0:
        pb = ProgressBar()
    else:
        devnull = DevNull()
        pb = ProgressBar(devnull)
    for _, (spin, a, l) in pb.enumerate(sal_i):
        symbol = chem_symbols[a]

        if soc:
            if world.rank == 0:  # GPAW soc is done in serial
                energies, weights = ldos(calc0, a, spin, l, theta, phi)
                mpi.broadcast((energies, weights))
            else:
                energies, weights = mpi.broadcast(None)
        else:
            energies, weights = ldos(calc, a, spin, l)

        # Reshape energies
        energies.shape = (kd.nibzkpts, -1)
        energies = energies[kd.bz2ibz_k]
        energies.shape = tuple(kd.N_c) + (-1, )

        # Get true weights and reshape
        weights.shape = (kd.nibzkpts, -1)
        weights /= kd.weight_k[:, np.newaxis]
        w = weights[kd.bz2ibz_k]
        w.shape = tuple(kd.N_c) + (-1, )

        # Linear tetrahedron integration
        p = lti(calc.atoms.cell, energies * Ha, e_e, w)

        # Store in dictionary
        key = ','.join([str(spin), str(symbol), str(l)])
        pdos_syl[key] += p

    return e_e, pdos_syl, calc.atoms.get_chemical_symbols(), efermi


def get_l_a(zs):
    """Define which atoms and angular momentum to project onto.

    Parameters
    ----------
    zs : [z1, z2, ...]-list or array
        list of atomic numbers (zi: int)

    Returns
    -------
    l_a : {int: str, ...}-dict
        keys are atomic indices and values are a string such as 'spd'
        that determines which angular momentum to project onto or a
        given atom

    """
    lantha = range(58, 72)
    acti = range(90, 104)

    zs = np.asarray(zs)
    l_a = {}
    atoms = Atoms(numbers=zs)
    mag_elements = magnetic_atoms(atoms)
    for a, (z, mag) in enumerate(zip(zs, mag_elements)):
        if z in lantha or z in acti:
            l_a[a] = 'spdf'
        else:
            l_a[a] = 'spd' if mag else 'sp'
    return l_a


# ----- DOS at Fermi energy ----- #


def dos_at_ef(calc, gpw, soc=True):
    """Get dos at the Fermi energy."""
    if soc:
        dos = SOCDOS(gpw, width=0.0, window=(-0.1, 0.1), npts=3)
    else:
        dos = DOS(calc, width=0.0, window=(-0.1, 0.1), npts=3)
    return dos.get_dos()[1]


# ---------- Plotting ---------- #


def get_ordered_syl_dict(dct_syl, symbols):
    """Order a dictionary with syl keys.

    Parameters
    ----------
    dct_syl : dict
        Dictionary with keys f'{s},{y},{l}'
        (spin (s), chemical symbol (y), angular momentum (l))
    symbols : list
        Sort symbols after index in this list

    Returns
    -------
    outdct_syl : OrderedDict
        Sorted dct_syl

    """
    from collections import OrderedDict

    # Setup ssili (spin, symbol index, angular momentum index) key
    def ssili(syl):
        s, y, L = syl.split(',')
        # Symbols list can have multiple entries of the same symbol
        # ex. ['O', 'Fe', 'O']. In this case 'O' will have index 0 and
        # 'Fe' will have index 1.
        si = symbols.index(y)
        li = ['s', 'p', 'd', 'f'].index(L)
        return f'{s}{si}{li}'

    return OrderedDict(sorted(dct_syl.items(), key=lambda t: ssili(t[0])))


def get_yl_colors(dct_syl):
    """Get the color indices corresponding to each symbol and angular momentum.

    Parameters
    ----------
    dct_syl : OrderedDict
        Ordered dictionary with keys f'{s},{y},{l}'
        (spin (s), chemical symbol (y), angular momentum (l))

    Returns
    -------
    color_yl : OrderedDict
        Color strings for each symbol and angular momentum

    """
    from collections import OrderedDict

    color_yl = OrderedDict()
    c = 0
    for key in dct_syl:
        # Do not differentiate spin by color
        if int(key[0]) == 0:  # if spin is 0
            color_yl[key[2:]] = 'C{}'.format(c)
            c += 1
            c = c % 10  # only 10 colors available in cycler

    return color_yl


def plot_pdos_nosoc(*args, **kwargs):
    return plot_pdos(*args, soc=False, **kwargs)


def plot_pdos_soc(*args, **kwargs):
    return plot_pdos(*args, soc=True, **kwargs)


def plot_pdos(row, filename, soc=True,
              figsize=(5.5, 5),
              lw=1, loc='best'):

    def smooth(y, npts=3):
        return np.convolve(y, np.ones(npts) / npts, mode='same')

    # Check if pdos data is stored in row
    results = 'results-asr.pdos.json'
    pdos = 'pdos_soc' if soc else 'pdos_nosoc'
    if results in row.data and pdos in row.data[results]:
        data = row.data[results][pdos]
    else:
        return

    import matplotlib.pyplot as plt
    import matplotlib.patheffects as path_effects

    # Extract raw data
    symbols = data['symbols']
    pdos_syl = get_ordered_syl_dict(data['pdos_syl'], symbols)
    e_e = data['energies'].copy() - row.get('evac', 0)
    ef = data['efermi']

    # Find energy range to plot in
    if soc:
        emin = row.get('vbm', ef) - 3 - row.get('evac', 0)
        emax = row.get('cbm', ef) + 3 - row.get('evac', 0)
    else:
        nosoc_data = row.data['results-asr.gs.json']['gaps_nosoc']
        vbmnosoc = nosoc_data.get('vbm', ef)
        cbmnosoc = nosoc_data.get('cbm', ef)

        if vbmnosoc is None:
            vbmnosoc = ef

        if cbmnosoc is None:
            cbmnosoc = ef

        emin = vbmnosoc - 3 - row.get('evac', 0)
        emax = cbmnosoc + 3 - row.get('evac', 0)

    # Set up energy range to plot in
    i1, i2 = abs(e_e - emin).argmin(), abs(e_e - emax).argmin()

    # Get color code
    color_yl = get_yl_colors(pdos_syl)

    # Figure out if pdos has been calculated for more than one spin channel
    spinpol = False
    for k in pdos_syl.keys():
        if int(k[0]) == 1:
            spinpol = True
            break

    # Set up plot
    plt.figure(figsize=figsize)
    ax = plt.gca()

    # Plot pdos
    pdosint_s = defaultdict(float)
    for key in pdos_syl:
        pdos = pdos_syl[key]
        spin, symbol, lstr = key.split(',')
        spin = int(spin)
        sign = 1 if spin == 0 else -1

        # Integrate pdos to find suiting pdos range
        pdosint_s[spin] += np.trapz(y=pdos[i1:i2], x=e_e[i1:i2])

        # Label atomic symbol and angular momentum
        if spin == 0:
            label = '{} ({})'.format(symbol, lstr)
        else:
            label = None

        ax.plot(smooth(pdos) * sign, e_e,
                label=label, color=color_yl[key[2:]])

    ax.legend(loc=loc)
    ax.axhline(ef - row.get('evac', 0), color='k', ls=':')

    # Set up axis limits
    ax.set_ylim(emin, emax)
    if spinpol:  # Use symmetric limits
        xmax = max(pdosint_s.values())
        ax.set_xlim(-xmax * 0.5, xmax * 0.5)
    else:
        ax.set_xlim(0, pdosint_s[0] * 0.5)

    # Annotate E_F
    xlim = ax.get_xlim()
    x0 = xlim[0] + (xlim[1] - xlim[0]) * 0.01
    text = plt.text(x0, ef - row.get('evac', 0),
                    r'$E_\mathrm{F}$',
                    ha='left',
                    va='bottom')

    text.set_path_effects([
        path_effects.Stroke(linewidth=3, foreground='white', alpha=0.5),
        path_effects.Normal()
    ])

    ax.set_xlabel('projected dos [states / eV]')
    if row.get('evac') is not None:
        ax.set_ylabel(r'$E-E_\mathrm{vac}$ [eV]')
    else:
        ax.set_ylabel(r'$E$ [eV]')

    plt.savefig(filename, bbox_inches='tight')
    plt.close()


if __name__ == '__main__':
    main.cli()



import pytest
from .conftest import test_materials


@pytest.mark.ci
def test_database_totree(separate_folder):
    from ase.db import connect
    from asr.database.totree import main
    from pathlib import Path

    dbname = 'database.db'
    db = connect(dbname)
    for atoms in test_materials:
        db.write(atoms)

    main(database=dbname)

    assert not Path('tree').is_dir()

    main(database=dbname, run=True)

    assert Path('tree').is_dir()
    assert Path('tree/A/123/Ag-A-123-c-0/structure.json').is_file()
    assert Path('tree/A/227/Si2-A-227-b-0/structure.json').is_file()
    assert Path('tree/AB/187/BN-AB-187-b-d-0/structure.json').is_file()



from asr.core import command, argument, option


@command('asr.setup.scanparams')
@argument('scanparams', nargs=-1,
          metavar='recipe:option arg arg arg recipe:option arg arg arg')
@option('--symlink/--no-symlink',
        help='Make symbolic link to everything '
        'in this folder (except params.json)')
def main(scanparams, symlink=True):
    """Make folders with different sets of parameters.

    This function will take a number of arguments in the syntax
    recipe:option arg arg arg and setup a number of folders named
    "scanparams*/" (where * is an integer) with a params.json file.

    If you set multiple options setup.scanparams will test all combinations
    of parameters.

    The function checks if any of the parameter combinations you ask for
    already exists and skips these parameter combinations if this is the case.

    This function can be useful when conducting convergence tests of recipes.

    Examples
    --------
    Test different kpoint density in the relax recipe
    >>> asr run "setup.scanparams asr.relax:kptdensity 3 4 5"
    Test combination of kpoint densities and planewave cutoff in relax:
    >>> asr run "setup.scanparams asr.relax:kptdensity 3 4 5 asr.relax:ecut 300 400 500"

    """
    from pathlib import Path
    from asr.core import get_recipes, read_json, write_json

    paramdict = {}
    recipes = get_recipes()
    for recipe in recipes:
        paramdict[recipe.name] = recipe.defparams

    # Find asr.recipe:option
    optioninds = []
    for i, arg in enumerate(scanparams):
        if isinstance(arg, str):
            if arg.startswith('asr.'):
                assert ':' in arg, 'You have to use the recipe:option syntax'
                optioninds.append(i)
    optioninds.append(len(scanparams))

    if Path('params.json').is_file():
        optionnames = [scanparams[ind] for ind in optioninds[:-1]]

        # Compile parameters from params.json
        orig_params = read_json('params.json')
        orignames = []
        for recipe, recipeparams in orig_params.items():
            for opt in recipeparams:
                orignames.append(f'{recipe}:{opt}')

        for opt in optionnames:
            assert opt not in orignames, \
                (f'You cannot set {opt} because it '
                 'is already set in params.json')

    splitscanparams = []
    scanparamsdict = {}
    for start, end in zip(optioninds[:-1], optioninds[1:]):
        params = scanparams[start:end]
        values = params[1:]
        assert ':' in params[0], 'You have to use the recipe:option syntax'
        assert len(values), 'You have to provide at least one parameter'
        recipe, option = params[0].split(':')
        assert recipe in paramdict, f'Unknown recipe: {recipe}'
        assert option in paramdict[recipe], \
            f'Unknown option: {recipe}:{option}'

        if recipe not in scanparamsdict:
            scanparamsdict[recipe] = {}
        assert option not in scanparamsdict[recipe], \
            'You cannot set {recipe}:{option} twice'

        newparams = [params[0]]
        for value in values:
            newparams.append(type(paramdict[recipe][option])(value))
        scanparamsdict[recipe][option] = newparams[1:]
        splitscanparams += [newparams]

    from itertools import product
    parameternames = [params[0] for params in splitscanparams]
    scanparams = [params[1:] for params in splitscanparams]
    allparams = []
    for values in product(*scanparams):
        params = {}
        for name, value in zip(parameternames, values):
            recipe, option = name.split(':')
            if recipe not in params:
                params[recipe] = {}
            assert option not in params[recipe], 'This should not happen!'
            params[recipe][option] = value
        allparams.append(params)

    # Find parameter combinations that have already been used
    newparams = []
    maxind = 0
    for p in Path().glob('scanparams*'):
        if not p.is_dir():
            continue
        assert (p / 'params.json').is_file(), \
            "{p}/params.json should exist but it doesn't"
        params = read_json(str(p / 'params.json'))

        for i, generatedparams in enumerate(allparams):
            if params == generatedparams:
                print(f'{params} already exists in {p}')
                allparams.pop(i)
                break
        maxind = max(maxind, int(str(p)[10:]) + 1)

    if not allparams:
        print('All parameter combinations already exists. '
              'Generated no new folders.')

    for j, params in enumerate(allparams):
        folder = Path(f'scanparams{maxind + j}')
        print(f'Generating {folder} with params: {params}')
        folder.mkdir()
        if symlink:
            for p in Path().glob('*'):
                if not p.is_file():
                    continue
                if p.name == 'params.json':
                    continue
                (folder / p.name).resolve().symlink_to(p.resolve())
        write_json(str(folder / 'params.json'), params)


if __name__ == '__main__':
    main.cli()



from asr.core import command, option


@command('asr.workflow')
@option("--include", help="Comma separated list of includes.")
@option("--exclude", help="Comma separated list of exludes.")
def main(include=None, exclude=None):
    """Run a full material property workflow."""
    import urllib.request
    from asr.core import get_recipes

    if include is not None:
        include = include.split(",")

    if exclude is not None:
        exclude = exclude.split(",")

    recipes = get_recipes()
    order = [
        "asr.relax",
        "asr.gs",
        "asr.convex_hull",
        "asr.phonons",
        "asr.setup_strains",
        "asr.magnetic_anisotropy",
        "asr.stiffness",
        "asr.emasses",
        "asr.pdos",
        "asr.bandstructure",
        "asr.projected_bandstructure",
        "asr.polarizability",
        "asr.plasmafrequency",
        "asr.fermisurface",
        "asr.borncharges",
        "asr.piezoelectrictensor"
        "asr.infrared_polarizability",
        "asr.push",
        "asr.raman",
        "asr.bse"
    ]
    url = ("https://cmr.fysik.dtu.dk/_downloads/"
           "ebe5e92dd4d83fd16999ce911c8527ab/oqmd12.db")
    urllib.request.urlretrieve(url, "oqmd12.db")

    extra_args = {"asr.convex_hull": {"databases": ["oqmd12.db"]}}

    def filterfunc(x):
        if exclude and x.name in exclude:
            return False
        if include and x.name not in include:
            return False
        if x.name not in order:
            return False
        return True

    recipes = filter(filterfunc, recipes)
    recipes = sorted(recipes, key=order.index)
    for recipe in recipes:
        kwargs = extra_args.get(recipe.name, {})
        recipe(**kwargs)


if __name__ == '__main__':
    main.cli()



from asr.core import command, option, read_json
from click import Choice


# This function is basically doing the exact same as HSE and could
# probably be refactored
def bs_gw(row,
          filename='gw-bs.png',
          figsize=(5.5, 5),
          fontsize=10,
          show_legend=True,
          s=0.5):
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import matplotlib.patheffects as path_effects

    data = row.data.get('results-asr.gw.json')
    path = data['bandstructure']['path']
    mpl.rcParams['font.size'] = fontsize
    ef = data['efermi_gw_soc']

    if row.get('evac') is not None:
        label = r'$E - E_\mathrm{vac}$ [eV]'
        reference = row.get('evac')
    else:
        label = r'$E - E_\mathrm{F}$ [eV]'
        reference = ef

    emin = row.get('vbm_gw', ef) - 3 - reference
    emax = row.get('cbm_gw', ef) + 3 - reference

    e_mk = data['bandstructure']['e_int_mk'] - reference
    x, X, labels = path.get_linear_kpoint_axis()

    # hse with soc
    style = dict(
        color='C1',
        ls='-',
        lw=1.0,
        zorder=0)
    ax = plt.figure(figsize=figsize).add_subplot(111)
    for e_m in e_mk:
        ax.plot(x, e_m, **style)
    ax.set_ylim([emin, emax])
    ax.set_xlim([x[0], x[-1]])
    ax.set_ylabel(label)
    ax.set_xlabel('$k$-points')
    ax.set_xticks(X)
    ax.set_xticklabels([lab.replace('G', r'$\Gamma$') for lab in labels])

    xlim = ax.get_xlim()
    x0 = xlim[1] * 0.01
    ax.axhline(ef - reference, c='C1', ls=':')
    text = ax.annotate(
        r'$E_\mathrm{F}$',
        xy=(x0, ef - reference),
        ha='left',
        va='bottom',
        fontsize=fontsize * 1.3)
    text.set_path_effects([
        path_effects.Stroke(linewidth=2, foreground='white', alpha=0.5),
        path_effects.Normal()
    ])

    # add PBE band structure with soc
    from asr.bandstructure import add_bs_pbe
    if 'results-asr.bandstructure.json' in row.data:
        ax = add_bs_pbe(row, ax, reference=row.get('evac', row.get('efermi')),
                        color=[0.8, 0.8, 0.8])

    for Xi in X:
        ax.axvline(Xi, ls='-', c='0.5', zorder=-20)

    ax.plot([], [], **style, label='GW')
    plt.legend(loc='upper right')

    if not show_legend:
        ax.legend_.remove()
    plt.savefig(filename, bbox_inches='tight')


def get_kpts_size(atoms, kptdensity):
    """Try to get a reasonable monkhorst size which hits high symmetry points."""
    from gpaw.kpt_descriptor import kpts2sizeandoffsets as k2so
    size, offset = k2so(atoms=atoms, density=kptdensity)
    size[2] = 1
    for i in range(2):
        if size[i] % 6 != 0:
            size[i] = 6 * (size[i] // 6 + 1)
    kpts = {'size': size, 'gamma': True}
    return kpts


@command(requires=['gs.gpw'],
         dependencies=['asr.gs@calculate'],
         creates=['gs_gw.gpw', 'gs_gw_nowfs.gpw'])
@option('--kptdensity', help='K-point density')
@option('--ecut', help='Plane wave cutoff')
def gs(kptdensity=5.0, ecut=200.0):
    """Calculate GW underlying ground state."""
    from ase.dft.bandgap import bandgap
    from gpaw import GPAW
    import numpy as np

    # check that the system is a semiconductor
    calc = GPAW('gs.gpw', txt=None)
    pbe_gap, _, _ = bandgap(calc, output=None)
    if pbe_gap < 0.05:
        raise Exception("GW: Only for semiconductors, PBE gap = "
                        + str(pbe_gap) + " eV is too small!")

    # check that the system is small enough
    atoms = calc.get_atoms()
    if len(atoms) > 4:
        raise Exception("GW: Only for small systems, "
                        + str(len(atoms)) + " > 4 atoms!")

    # setup k points/parameters
    dim = np.sum(atoms.pbc.tolist())
    if dim == 3:
        kpts = {'density': kptdensity, 'gamma': True, 'even': True}
    elif dim == 2:
        kpts = get_kpts_size(atoms=atoms, kptdensity=kptdensity)
    elif dim == 1:
        kpts = {'density': kptdensity, 'gamma': True, 'even': True}
        # TODO remove unnecessary k
        raise NotImplementedError('asr for dim=1 not implemented!')
    elif dim == 0:
        kpts = {'density': 0.0, 'gamma': True, 'even': True}
        # TODO only Gamma
        raise NotImplementedError('asr for dim=0 not implemented!')

    # we need energies/wavefunctions on the correct grid
    calc = GPAW(
        'gs.gpw',
        txt='gs_gw.txt',
        fixdensity=True,
        kpts=kpts,
        parallel={'domain': 1})
    calc.get_potential_energy()
    calc.diagonalize_full_hamiltonian(ecut=ecut)
    calc.write('gs_gw_nowfs.gpw')
    calc.write('gs_gw.gpw', mode='all')


@command(requires=['gs_gw.gpw'],
         dependencies=['asr.gw@gs'])
@option('--ecut', help='Plane wave cutoff')
@option('--mode', help='GW mode',
        type=Choice(['G0W0', 'GWG']))
def gw(ecut=200.0, mode='G0W0'):
    """Calculate GW corrections."""
    from ase.dft.bandgap import bandgap
    from gpaw import GPAW
    from gpaw.response.g0w0 import G0W0
    import numpy as np

    # check that the system is a semiconductor
    calc = GPAW('gs.gpw', txt=None)
    pbe_gap, _, _ = bandgap(calc, output=None)
    if pbe_gap < 0.05:
        raise Exception("GW: Only for semiconductors, PBE gap = "
                        + str(pbe_gap) + " eV is too small!")

    # check that the system is small enough
    atoms = calc.get_atoms()
    if len(atoms) > 4:
        raise Exception("GW: Only for small systems, "
                        + str(len(atoms)) + " > 4 atoms!")

    # Setup parameters
    dim = np.sum(atoms.pbc.tolist())
    if dim == 3:
        truncation = 'wigner-seitz'
        q0_correction = False
    elif dim == 2:
        truncation = '2D'
        q0_correction = True
    elif dim == 1:
        raise NotImplementedError('asr for dim=1 not implemented!')
        truncation = '1D'
        q0_correction = False
    elif dim == 0:
        raise NotImplementedError('asr for dim=0 not implemented!')
        truncation = '0D'
        q0_correction = False

    if mode == 'GWG':
        raise NotImplementedError('GW: asr for GWG not implemented!')

    lb, ub = max(calc.wfs.nvalence // 2 - 8, 0), calc.wfs.nvalence // 2 + 4

    calc = G0W0(calc='gs_gw.gpw',
                bands=(lb, ub),
                ecut=ecut,
                ecut_extrapolation=True,
                truncation=truncation,
                nblocksmax=True,
                q0_correction=q0_correction,
                filename='g0w0',
                restartfile='g0w0.tmp',
                savepckl=False)

    results = calc.calculate()
    results['minband'] = lb
    results['maxband'] = ub
    return results


def webpanel(row, key_descriptions):
    from asr.database.browser import fig, table
    if not row.get('gap_gw', 0) > 0:
        return []

    prop = table(row, 'Property', [
        'gap_gw', 'gap_dir_gw',
    ], key_descriptions)

    if row.get('evac'):
        prop['rows'].extend(
            [['Valence band maximum wrt. vacuum level (G0W0)',
              f'{row.vbm_gw - row.evac:.2f} eV'],
             ['Conduction band minimum wrt. vacuum level (G0W0)',
              f'{row.cbm_gw - row.evac:.2f} eV']])
    else:
        prop['rows'].extend(
            [['Valence band maximum wrt. Fermi level (G0W0)',
              f'{row.vbm_gw - row.efermi:.2f} eV'],
             ['Conduction band minimum wrt. Fermi level (G0W0)',
              f'{row.cbm_gw - row.efermi:.2f} eV']])

    panel = {'title': 'Electronic band structure (GW)',
             'columns': [[fig('gw-bs.png')], [prop]],
             'plot_descriptions': [{'function': bs_gw,
                                    'filenames': ['gw-bs.png']}],
             'sort': 16}

    if row.get('gap_gw'):
        rows = [['Band gap (G0W0)', f'{row.gap_gw:0.2f} eV']]

        summary = {'title': 'Summary',
                   'columns': [[{'type': 'table',
                                 'header': ['Electronic properties', ''],
                                 'rows': rows}]],
                   'sort': 12}

        return [panel, summary]

    return [panel]


@command(requires=['results-asr.gw@gw.json', 'gs_gw_nowfs.gpw',
                   'results-asr.bandstructure.json'],
         dependencies=['asr.gw@gw', 'asr.gw@gs', 'asr.bandstructure'],
         webpanel=webpanel)
def main():
    import numpy as np
    from gpaw import GPAW
    from asr.utils.gpw2eigs import fermi_level
    from ase.dft.bandgap import bandgap
    from asr.hse import MP_interpolate
    from types import SimpleNamespace

    calc = GPAW('gs_gw_nowfs.gpw', txt=None)
    gwresults = SimpleNamespace(**read_json('results-asr.gw@gw.json'))

    lb = gwresults.minband
    ub = gwresults.maxband

    delta_skn = gwresults.qp - gwresults.eps

    # Interpolate band structure
    results = MP_interpolate(calc, delta_skn, lb, ub)
    kd = {}

    # First get stuff without SOC
    eps_skn = gwresults.qp
    efermi_nosoc = fermi_level(calc, eps_skn=eps_skn,
                               nelectrons=(calc.get_number_of_electrons()
                                           - 2 * lb))
    gap, p1, p2 = bandgap(eigenvalues=eps_skn, efermi=efermi_nosoc,
                          output=None)
    gapd, p1d, p2d = bandgap(eigenvalues=eps_skn, efermi=efermi_nosoc,
                             direct=True, output=None)
    if gap > 0:
        ibzkpts = calc.get_ibz_k_points()
        kvbm_nosoc = ibzkpts[p1[1]]  # k coordinates of vbm
        kcbm_nosoc = ibzkpts[p2[1]]  # k coordinates of cbm
        vbm = eps_skn[p1]
        cbm = eps_skn[p2]
        subresults = {'vbm_gw_nosoc': vbm,
                      'cbm_gw_nosoc': cbm,
                      'gap_dir_gw_nosoc': gapd,
                      'gap_gw_nosoc': gap,
                      'kvbm_nosoc': kvbm_nosoc,
                      'kcbm_nosoc': kcbm_nosoc}

        kd.update({'vbm_gw_nosoc': 'GW valence band max. w/o soc [eV]',
                   'cbm_gw_nosoc': 'GW condution band min. w/o soc [eV]',
                   'gap_dir_gw_nosoc': 'GW direct gap w/o soc [eV]',
                   'gap_gw_nosoc': 'GW gap w/o soc [eV]',
                   'kvbm_nosoc': 'k-point of GW valence band max. w/o soc',
                   'kcbm_nosoc': 'k-point of GW conduction band min. w/o soc'})
        results.update(subresults)

    # Get the SO corrected GW QP energires
    from gpaw.spinorbit import get_spinorbit_eigenvalues as get_soc_eigs
    from asr.magnetic_anisotropy import get_spin_axis
    bandrange = np.arange(lb, ub)
    theta, phi = get_spin_axis()
    e_mk = get_soc_eigs(calc, gw_kn=eps_skn,
                        bands=bandrange,
                        return_spin=False,
                        theta=theta, phi=phi)

    eps = e_mk.transpose()[np.newaxis]  # e_skm, dummy spin index
    efermi_soc = fermi_level(calc, eps_skn=eps,
                             nelectrons=(2 * (calc.get_number_of_electrons()
                                              - 2 * lb)))
    gap, p1, p2 = bandgap(eigenvalues=eps, efermi=efermi_soc,
                          output=None)
    gapd, p1d, p2d = bandgap(eigenvalues=eps, efermi=efermi_soc,
                             direct=True, output=None)
    if gap:
        kvbm = ibzkpts[p1[1]]
        kcbm = ibzkpts[p2[1]]
        vbm = eps[p1]
        cbm = eps[p2]
        subresults = {'vbm_gw': vbm,
                      'cbm_gw': cbm,
                      'gap_dir_gw': gapd,
                      'gap_gw': gap,
                      'kvbm': kvbm,
                      'kcbm': kcbm}
        kd.update({'vbm_gw': 'KVP: GW valence band max. [eV]',
                   'cbm_gw': 'KVP: GW conduction band min. [eV]',
                   'gap_dir_gw': 'KVP: GW direct gap [eV]',
                   'gap_gw': 'KVP: GW gap [eV]',
                   'kvbm': 'k-point of GW valence band max.',
                   'kcbm': 'k-point of GW conduction band min.'})
        results.update(subresults)

    results.update({'efermi_gw_nosoc': efermi_nosoc,
                    'efermi_gw_soc': efermi_soc})
    kd.update({'efermi_gw_nosoc': 'GW Fermi energy w/o soc [eV]',
               'efermi_gw_soc': 'GW Fermi energy [eV]'})
    results['__key_descriptions__'] = kd

    return results


if __name__ == '__main__':
    main.cli()



from asr.core import command, option


@command('asr.push',
         dependencies=['asr.structureinfo', 'asr.phonons'])
@option('-q', '--momentum', nargs=3, type=float,
        help='Phonon momentum')
@option('-m', '--mode', type=int, help='Mode index')
@option('-a', '--amplitude', type=float,
        help='Maximum distance an atom will be displaced')
def main(momentum=[0, 0, 0], mode=0, amplitude=0.1):
    """Push structure along some phonon mode and relax structure."""
    from asr.phonons import analyse
    import numpy as np
    q_c = momentum

    # Get modes
    from ase.io import read
    atoms = read('structure.json')
    omega_kl, u_klav, q_qc = analyse(modes=True, q_qc=[q_c])

    # Repeat atoms
    from fractions import Fraction
    repeat_c = [Fraction(qc).denominator if qc > 1e-3 else 1 for qc in q_qc[0]]
    newatoms = atoms * repeat_c

    # Here ``Na`` refers to a composite unit cell/atom dimension
    pos_Nav = newatoms.get_positions()

    # Total number of unit cells
    N = np.prod(repeat_c)

    # Corresponding lattice vectors R_m
    R_cN = np.indices(repeat_c).reshape(3, -1)

    # Bloch phase
    phase_N = np.exp(2j * np.pi * np.dot(q_c, R_cN))
    phase_Na = phase_N.repeat(len(atoms))

    # Repeat and multiply by Bloch phase factor
    mode_av = u_klav[0, mode]
    n_a = np.linalg.norm(mode_av, axis=1)
    mode_av /= np.max(n_a)
    mode_Nav = (np.vstack(N * [mode_av]) * phase_Na[:, np.newaxis] * amplitude)
    newatoms.set_positions(pos_Nav + mode_Nav.real)

    # Write unrelaxed.json file to folder
    folder = 'push-q-{}-{}-{}-mode-{}'.format(q_c[0], q_c[1], q_c[2], mode)
    from gpaw.mpi import world
    from pathlib import Path
    from ase.io import write
    if world.rank == 0 and not Path(folder).is_dir():
        Path(folder).mkdir()
    write(f'{folder}/unrelaxed.json', newatoms)


if __name__ == '__main__':
    main.cli()



from pathlib import Path

import numpy as np

from ase.parallel import world
from ase.io import read
from ase.phonons import Phonons

from asr.core import command, option


def creates():
    atoms = read('structure.json')
    natoms = len(atoms)
    filenames = ['phonon.eq.pckl']
    for a in range(natoms):
        for v in 'xyz':
            for pm in '+-':
                # Atomic forces for a displacement of atom a in direction v
                filenames.append(f'phonon.{a}{v}{pm}.pckl')
    return filenames


def todict(filename):
    from ase.utils import pickleload
    return {'content': pickleload(open(filename, 'rb'))}


def topckl(filename, dct):
    from ase.utils import opencew
    import pickle
    contents = dct['content']
    fd = opencew(filename)
    if world.rank == 0:
        pickle.dump(contents, fd, protocol=2)
        fd.close()


@command('asr.phonons',
         requires=['structure.json', 'gs.gpw'],
         dependencies=['asr.gs@calculate'],
         creates=creates)
@option('-n', help='Supercell size')
@option('--ecut', help='Energy cutoff')
@option('--kptdensity', help='Kpoint density')
@option('--fconverge', help='Force convergence criterium')
def calculate(n=2, ecut=800, kptdensity=6.0, fconverge=1e-4):
    """Calculate atomic forces used for phonon spectrum."""
    from asr.calculators import get_calculator
    # Remove empty files:
    if world.rank == 0:
        for f in Path().glob('phonon.*.pckl'):
            if f.stat().st_size == 0:
                f.unlink()
    world.barrier()

    atoms = read('structure.json')
    gsold = get_calculator()('gs.gpw', txt=None)

    # Set initial magnetic moments
    from asr.core import is_magnetic
    if is_magnetic():
        magmoms_m = gsold.get_magnetic_moments()
        atoms.set_initial_magnetic_moments(magmoms_m)

    params = gsold.parameters.copy()  # TODO: remove fix density from gs params
    if 'fixdensity' in params:
        params.pop('fixdensity')
    params.update({'mode': {'name': 'pw', 'ecut': ecut},
                   'kpts': {'density': kptdensity, 'gamma': True}})

    # Set essential parameters for phonons
    params['symmetry'] = {'point_group': False}

    # Make sure to converge forces! Can be important
    params['convergence'] = {'forces': fconverge}

    fd = open('phonons.txt'.format(n), 'a')
    params['txt'] = fd
    calc = get_calculator()(**params)

    from asr.core import get_dimensionality
    nd = get_dimensionality()
    if nd == 3:
        supercell = (n, n, n)
    elif nd == 2:
        supercell = (n, n, 1)
    elif nd == 1:
        supercell = (n, 1, 1)

    p = Phonons(atoms=atoms, calc=calc, supercell=supercell)
    p.run()

    # Read creates files
    files = {}
    for filename in creates():
        dct = todict(filename)
        dct['__tofile__'] = 'asr.phonons@topckl'
        files[filename] = dct
    data = {'__files__': files}
    fd.close()
    return data


def requires():
    return creates() + ['results-asr.phonons@calculate.json']


def webpanel(row, key_descriptions):
    from asr.database.browser import table, fig
    phonontable = table(row, 'Property', ['minhessianeig'], key_descriptions)

    panel = {'title': 'Phonons',
             'columns': [[fig('phonon_bs.png')], [phonontable]],
             'plot_descriptions': [{'function': plot_bandstructure,
                                    'filenames': ['phonon_bs.png']}],
             'sort': 3}

    dynstab = row.get('dynamic_stability_level')
    stabilities = {1: 'low', 2: 'medium', 3: 'high'}
    high = 'Min. Hessian eig. > -0.01 meV/Ang^2'
    medium = 'Min. Hessian eig. > -2 eV/Ang^2'
    low = 'Min. Hessian eig.  < -2 eV/Ang^2'
    row = ['Dynamical (phonons)',
           '<a href="#" data-toggle="tooltip" data-html="true" '
           + 'title="LOW: {}&#13;MEDIUM: {}&#13;HIGH: {}">{}</a>'.format(
               low, medium, high, stabilities[dynstab].upper())]

    summary = {'title': 'Summary',
               'columns': [[{'type': 'table',
                             'header': ['Stability', 'Category'],
                             'rows': [row]}]],
               'sort': 2}
    return [panel, summary]


@command('asr.phonons',
         requires=requires,
         webpanel=webpanel,
         dependencies=['asr.phonons@calculate'])
@option('--mingo/--no-mingo', is_flag=True,
        help='Perform Mingo correction of force constant matrix')
def main(mingo=True):
    from asr.core import read_json
    from asr.core import get_dimensionality
    dct = read_json('results-asr.phonons@calculate.json')
    atoms = read('structure.json')
    n = dct['__params__']['n']
    nd = get_dimensionality()
    if nd == 3:
        supercell = (n, n, n)
    elif nd == 2:
        supercell = (n, n, 1)
    elif nd == 1:
        supercell = (n, 1, 1)
    p = Phonons(atoms=atoms, supercell=supercell)
    p.read(symmetrize=0)

    if mingo:
        # We correct the force constant matrix and
        # dynamical matrix
        C_N = mingocorrection(p.C_N, atoms, supercell)
        p.C_N = C_N

        # Calculate dynamical matrix
        D_N = C_N.copy()
        m_a = atoms.get_masses()
        m_inv_x = np.repeat(m_a**-0.5, 3)
        M_inv = np.outer(m_inv_x, m_inv_x)
        for D in D_N:
            D *= M_inv
            p.D_N = D_N

    # First calculate the exactly known q-points
    q_qc = np.indices(p.N_c).reshape(3, -1).T / p.N_c
    out = p.band_structure(q_qc, modes=True, born=False, verbose=False)
    omega_kl, u_kl = out

    R_cN = p.lattice_vectors()
    eigs = []
    for q_c in q_qc:
        phase_N = np.exp(-2j * np.pi * np.dot(q_c, R_cN))
        C_q = np.sum(phase_N[:, np.newaxis, np.newaxis] * p.C_N, axis=0)
        eigs.append(np.linalg.eigvalsh(C_q))

    eigs = np.array(eigs)
    mineig = np.min(eigs)

    if mineig < -2:
        dynamic_stability = 1
    elif mineig < -1e-5:
        dynamic_stability = 2
    else:
        dynamic_stability = 3

    results = {'omega_kl': omega_kl,
               'q_qc': q_qc,
               'modes_kl': u_kl,
               'minhessianeig': mineig,
               'dynamic_stability_level': dynamic_stability}

    # Next calculate an approximate phonon band structure
    path = atoms.cell.bandpath(npoints=100, pbc=atoms.pbc)
    freqs_kl = p.band_structure(path.kpts, modes=False, born=False,
                                verbose=False)
    results['interp_freqs_kl'] = freqs_kl
    results['path'] = path
    results['__key_descriptions__'] = \
        {'minhessianeig': 'KVP: Minimum eigenvalue of Hessian [eV/Ang^2]',
         'dynamic_stability_level': 'KVP: Dynamic stability level'}

    return results


def plot_phonons(row, fname):
    import matplotlib.pyplot as plt

    data = row.data.get('results-asr.phonons.json')
    if data is None:
        return

    omega_kl = data['omega_kl']
    gamma = omega_kl[0]
    fig = plt.figure(figsize=(6.4, 3.9))
    ax = fig.gca()

    x0 = -0.0005  # eV
    for x, color in [(gamma[gamma < x0], 'r'),
                     (gamma[gamma >= x0], 'b')]:
        if len(x) > 0:
            markerline, _, _ = ax.stem(x * 1000, np.ones_like(x), bottom=-1,
                                       markerfmt=color + 'o',
                                       linefmt=color + '-')
            plt.setp(markerline, alpha=0.4)
    ax.set_xlabel(r'phonon frequency at $\Gamma$ [meV]')
    ax.axis(ymin=0.0, ymax=1.3)
    plt.tight_layout()
    plt.savefig(fname)
    plt.close()


def plot_bandstructure(row, fname):
    from matplotlib import pyplot as plt
    from ase.dft.band_structure import BandStructure
    data = row.data.get('results-asr.phonons.json')
    path = data['path']
    energies = data['interp_freqs_kl'] * 1e3
    exact_indices = []
    for q_c in data['q_qc']:
        diff_kc = path.kpts - q_c
        diff_kc -= np.round(diff_kc)
        inds = np.argwhere(np.all(np.abs(diff_kc) < 1e-3, 1))
        exact_indices.extend(inds.tolist())

    en_exact = np.zeros_like(energies) + np.nan
    for ind in exact_indices:
        en_exact[ind] = energies[ind]

    bs = BandStructure(path=path, energies=en_exact[None])
    bs.plot(ax=plt.gca(), ls='', marker='o', colors=['C0'],
            emin=np.min(energies * 1.1), emax=np.max(energies * 1.15),
            ylabel='Phonon frequencies [meV]')
    plt.plot([], [], label='Calculated', color='C1', marker='o', ls='')
    plt.legend(ncol=1, loc='upper center')
    plt.tight_layout()
    plt.savefig(fname)


def mingocorrection(Cin_NVV, atoms, supercell):
    na = len(atoms)
    nc = np.prod(supercell)
    dimension = nc * na * 3

    Cin = (Cin_NVV.reshape(*supercell, na, 3, na, 3).
           transpose(3, 4, 0, 1, 2, 5, 6))

    C = np.empty((*supercell, na, 3, *supercell, na, 3))

    from itertools import product
    for n1, n2, n3 in product(range(supercell[0]),
                              range(supercell[1]),
                              range(supercell[2])):
        inds1 = (np.arange(supercell[0]) - n1) % supercell[0]
        inds2 = (np.arange(supercell[1]) - n2) % supercell[1]
        inds3 = (np.arange(supercell[2]) - n3) % supercell[2]
        C[n1, n2, n3] = Cin[:, :, inds1][:, :, :, inds2][:, :, :, :, inds3]

    C.shape = (dimension, dimension)
    C += C.T.copy()
    C *= 0.5

    # Mingo correction.
    #
    # See:
    #
    #    Phonon transmission through defects in carbon nanotubes
    #    from first principles
    #
    #    N. Mingo, D. A. Stewart, D. A. Broido, and D. Srivastava
    #    Phys. Rev. B 77, 033418 – Published 30 January 2008
    #    http://dx.doi.org/10.1103/PhysRevB.77.033418

    R_in = np.zeros((dimension, 3))
    for n in range(3):
        R_in[n::3, n] = 1.0
    a_in = -np.dot(C, R_in)
    B_inin = np.zeros((dimension, 3, dimension, 3))
    for i in range(dimension):
        B_inin[i, :, i] = np.dot(R_in.T, C[i, :, np.newaxis]**2 * R_in) / 4
        for j in range(dimension):
            B_inin[i, :, j] += np.outer(R_in[i], R_in[j]).T * C[i, j]**2 / 4

    L_in = np.dot(np.linalg.pinv(B_inin.reshape((dimension * 3,
                                                 dimension * 3))),
                  a_in.reshape((dimension * 3,))).reshape((dimension, 3))
    D_ii = C**2 * (np.dot(L_in, R_in.T) + np.dot(L_in, R_in.T).T) / 4
    C += D_ii

    C.shape = (*supercell, na, 3, *supercell, na, 3)
    Cout = C[0, 0, 0].transpose(2, 3, 4, 0, 1, 5, 6).reshape(nc,
                                                             na * 3,
                                                             na * 3)
    return Cout


if __name__ == '__main__':
    main.cli()



from asr.core import command


@command('asr.bader',
         dependencies=['asr.structureinfo', 'asr.gs'])
def main():
    """Calculate bader charges.

    To make Bader analysis we use another program. Download the executable
    for Bader analysis and put in path (this is for Linux, find the
    appropriate executable for you own OS)

        $ mkdir baderext && cd baderext
        $ wget theory.cm.utexas.edu/henkelman/code/bader/download/
        ...bader_lnx_64.tar.gz
        $ tar -zxf bader_lnx_64.tar.gz
        $ echo  'export PATH=~/baderext:$PATH' >> ~/.bashrc
    """
    from pathlib import Path
    from ase.io import write
    from ase.units import Bohr
    from gpaw import GPAW
    from gpaw.mpi import world

    assert world.size == 1, print('Do not run in parallel!')

    gs = GPAW('gs.gpw', txt=None)
    atoms = gs.atoms
    density = gs.get_all_electron_density() * Bohr**3
    write('density.cube', atoms, data=density)

    import subprocess
    import os
    folder = 'data-bader'
    try:
        os.makedirs(folder)
    except FileExistsError:
        pass
    cmd = 'bader -p all_atom -p atom_index ../density.cube'
    out = (Path(folder) / 'bader.out').open('w')
    err = (Path(folder) / 'bader.err').open('w')
    subprocess.run(cmd.split(), cwd=folder,
                   stdout=out,
                   stderr=err)
    out.close()
    err.close()


def print():
    """Print Bader charges."""
    import os.path as op
    fname = 'data-bader/ACF.dat'
    if not op.isfile(fname):
        return

    with open(fname) as f:
        dat = f.read()
    print(dat)


if __name__ == '__main__':
    main.cli()



from asr.core import command, option
from click import Choice


def webpanel(row, key_descriptions):
    from asr.database.browser import fig, table

    opt = table(row, 'Property', [
        'alphax_el', 'alphay_el', 'alphaz_el',
        'alphax_lat', 'alphay_lat', 'alphaz_lat',
        'alphax', 'alphay', 'alphaz',
    ], key_descriptions)

    panel = {'title': 'Optical polarizability (RPA)',
             'columns': [[fig('rpa-pol-x.png'), fig('rpa-pol-z.png')],
                         [fig('rpa-pol-y.png'), opt]],
             'plot_descriptions':
                 [{'function': polarizability,
                   'filenames': ['rpa-pol-x.png',
                                 'rpa-pol-y.png',
                                 'rpa-pol-z.png']}],
             'sort': 20}

    return [panel]


def get_kpts_size(atoms, density):
    """Try to get a reasonable monkhorst size which hits high symmetry points."""
    from gpaw.kpt_descriptor import kpts2sizeandoffsets as k2so
    size, offset = k2so(atoms=atoms, density=density)
    size[2] = 1
    for i in range(2):
        if size[i] % 6 != 0:
            size[i] = 6 * (size[i] // 6 + 1)
    kpts = {'size': size, 'gamma': True}
    return kpts


@command('asr.polarizability',
         dependencies=['asr.structureinfo', 'asr.gs@calculate'],
         requires=['gs.gpw'],
         webpanel=webpanel)
@option(
    '--gs', help='Ground state on which response is based')
@option('--kptdensity', help='K-point density')
@option('--ecut', help='Plane wave cutoff')
@option('--xc', help='XC interaction', type=Choice(['RPA', 'ALDA']))
@option('--bandfactor', type=int,
        help='Number of unoccupied bands = (#occ. bands) * bandfactor)')
def main(gs='gs.gpw', kptdensity=20.0, ecut=50.0, xc='RPA', bandfactor=5):
    """Calculate linear response polarizability or dielectricfunction (only in 3D)."""
    from ase.io import read
    from gpaw import GPAW
    from gpaw.mpi import world
    from gpaw.response.df import DielectricFunction
    from pathlib import Path
    import numpy as np

    atoms = read('structure.json')
    pbc = atoms.pbc.tolist()

    dfkwargs = {
        'eta': 0.05,
        'domega0': 0.005,
        'ecut': ecut,
        'name': 'chi',
        'intraband': False
    }

    ND = np.sum(pbc)
    if ND == 3 or ND == 1:
        kpts = {'density': kptdensity, 'gamma': False, 'even': True}
    elif ND == 2:
        kpts = get_kpts_size(atoms=atoms, density=kptdensity)
        volume = atoms.get_volume()
        if volume < 120:
            nblocks = world.size // 4
        else:
            nblocks = world.size // 2
        dfkwargs.update({
            'nblocks': nblocks,
            'pbc': pbc,
            'integrationmode': 'tetrahedron integration',
            'truncation': '2D'
        })

    else:
        raise NotImplementedError(
            'Polarizability not implemented for 1D and 2D structures')

    if not Path('es.gpw').is_file():
        calc_old = GPAW(gs, txt=None)
        nval = calc_old.wfs.nvalence

        calc = GPAW(
            gs,
            txt='es.txt',
            fixdensity=True,
            nbands=(bandfactor + 1) * nval,
            convergence={'bands': bandfactor * nval},
            occupations={'name': 'fermi-dirac',
                         'width': 1e-4},
            kpts=kpts)
        calc.get_potential_energy()
        calc.write('es.gpw', mode='all')

    df = DielectricFunction('es.gpw', **dfkwargs)
    alpha0x, alphax = df.get_polarizability(
        q_c=[0, 0, 0], direction='x', pbc=pbc, filename=None,
        xc=xc)
    alpha0y, alphay = df.get_polarizability(
        q_c=[0, 0, 0], direction='y', pbc=pbc, filename=None,
        xc=xc)
    alpha0z, alphaz = df.get_polarizability(
        q_c=[0, 0, 0], direction='z', pbc=pbc, filename=None,
        xc=xc)

    plasmafreq_vv = df.chi0.plasmafreq_vv

    frequencies = df.get_frequencies()
    data = {
        'alpha0x_w': np.array(alpha0x),
        'alphax_w': np.array(alphax),
        'alpha0y_w': np.array(alpha0y),
        'alphay_w': np.array(alphay),
        'alpha0z_w': np.array(alpha0z),
        'alphaz_w': np.array(alphaz),
        'plasmafreq_vv': plasmafreq_vv,
        'frequencies': frequencies
    }

    data['alphax_el'] = data['alphax_w'][0].real
    data['alphay_el'] = data['alphay_w'][0].real
    data['alphaz_el'] = data['alphaz_w'][0].real

    data['__key_descriptions__'] = {
        'alphax_el': 'KVP: Static electronic polarizability,'
        ' x-direction [Ang]',
        'alphay_el': 'KVP: Static electronic polarizability,'
        ' y-direction [Ang]',
        'alphaz_el': 'KVP: Static electronic polarizability,'
        ' z-direction [Ang]'}

    world.barrier()
    if world.rank == 0:
        es_file = Path("es.gpw")
        es_file.unlink()

    return data


def polarizability(row, fx, fy, fz):
    import numpy as np
    import matplotlib.pyplot as plt

    def xlim():
        return (0, 10)

    def ylims(ws, data, wstart=0.0):
        i = abs(ws - wstart).argmin()
        x = data[i:]
        x1, x2 = x.real, x.imag
        y1 = min(x1.min(), x2.min()) * 1.02
        y2 = max(x1.max(), x2.max()) * 1.02
        return y1, y2

    data = row.data.get('results-asr.polarizability.json')

    if data is None:
        return
    frequencies = data['frequencies']
    i2 = abs(frequencies - 50.0).argmin()
    frequencies = frequencies[:i2]
    alphax_w = data['alphax_w'][:i2]
    alphay_w = data['alphay_w'][:i2]
    alphaz_w = data['alphaz_w'][:i2]

    infrared = row.data.get('results-asr.infraredpolarizability.json')
    if infrared:
        from scipy.interpolate import interp1d
        omegatmp_w = infrared['omega_w']
        alpha_wvv = infrared['alpha_wvv']
        alphax = interp1d(omegatmp_w, alpha_wvv[:, 0, 0], fill_value=0,
                          bounds_error=False)
        alphax_w = (alphax_w + alphax(frequencies))
        alphay = interp1d(omegatmp_w, alpha_wvv[:, 1, 1], fill_value=0,
                          bounds_error=False)
        alphay_w = (alphay_w + alphay(frequencies))
        alphaz = interp1d(omegatmp_w, alpha_wvv[:, 2, 2], fill_value=0,
                          bounds_error=False)
        alphaz_w = (alphaz_w + alphaz(frequencies))

    ax = plt.figure().add_subplot(111)
    ax1 = ax
    try:
        wpx = row.plasmafrequency_x
        if wpx > 0.01:
            alphaxfull_w = alphax_w - wpx**2 / (2 * np.pi * (frequencies + 1e-9)**2)
            ax.plot(
                frequencies,
                np.real(alphaxfull_w),
                '-',
                c='C1',
                label='real')
            ax.plot(
                frequencies,
                np.real(alphax_w),
                '--',
                c='C1',
                label='real interband')
        else:
            ax.plot(frequencies, np.real(alphax_w), c='C1', label='real')
    except AttributeError:
        ax.plot(frequencies, np.real(alphax_w), c='C1', label='real')
    ax.plot(frequencies, np.imag(alphax_w), c='C0', label='imag')
    ax.set_title('x-polarization')
    ax.set_xlabel('energy [eV]')
    ax.set_ylabel(r'Polarizability [$\mathrm{\AA}$]')
    ax.set_ylim(ylims(ws=frequencies, data=alphax_w, wstart=0.5))
    ax.legend()
    ax.set_xlim(xlim())
    plt.tight_layout()
    plt.savefig(fx)

    ax = plt.figure().add_subplot(111)
    ax2 = ax
    try:
        wpy = row.plasmafrequency_y
        if wpy > 0.01:
            alphayfull_w = alphay_w - wpy**2 / (2 * np.pi * (frequencies + 1e-9)**2)
            ax.plot(
                frequencies,
                np.real(alphayfull_w),
                '-',
                c='C1',
                label='real')
            ax.plot(
                frequencies,
                np.real(alphay_w),
                '--',
                c='C1',
                label='real interband')
        else:
            ax.plot(frequencies, np.real(alphay_w), c='C1', label='real')
    except AttributeError:
        ax.plot(frequencies, np.real(alphay_w), c='C1', label='real')
    ax.plot(frequencies, np.imag(alphay_w), c='C0', label='imag')
    ax.set_title('y-polarization')
    ax.set_xlabel('energy [eV]')
    ax.set_ylabel(r'Polarizability [$\mathrm{\AA}$]')
    ax.set_ylim(ylims(ws=frequencies, data=alphax_w, wstart=0.5))
    ax.legend()
    ax.set_xlim(xlim())
    plt.tight_layout()
    plt.savefig(fy)

    ax = plt.figure().add_subplot(111)
    ax3 = ax
    ax.plot(frequencies, np.real(alphaz_w), c='C1', label='real')
    ax.plot(frequencies, np.imag(alphaz_w), c='C0', label='imag')
    ax.set_title('z-polarization')
    ax.set_xlabel('energy [eV]')
    ax.set_ylabel(r'Polarizability [$\mathrm{\AA}$]')
    ax.set_ylim(ylims(ws=frequencies, data=alphaz_w, wstart=0.5))
    ax.legend()
    ax.set_xlim(xlim())
    plt.tight_layout()
    plt.savefig(fz)

    return ax1, ax2, ax3


if __name__ == '__main__':
    main.cli()
